from typing import Union, List

class ScreeningLogic:
    def __init__(
        self,
        id: int,
        variable_name: str,
        yyyymm: str,
        quantile: float,
        keep: str
    ):
        self.id = id
        self.variable_name = variable_name
        self.yyyymm = yyyymm
        if (quantile <= 0) or (quantile >= 1):
            raise Exception('Invalid quantile. Try to input a float in range (0,1).')
        self.quantile = quantile
        keep_set = {'bottom', 'top'}
        if keep not in keep_set:
            raise Exception(f'Invalid keep. Try to input one of {keep_set}.')
        self.keep = keep

    def __repr__(self):
        return f"<{self.__class__.__name__} ({self.id}): {self.variable_name}>"


class Portfolio:
    def __init__(
        self,
        id: int,
        date: str,
        stock_codes: List[str],
        weights: Union[List[float], None] = None,
        holding_period_in_month: Union[int, None] = None,
        label: Union[str, None] = None,
    ):
        self.id = id
        self.date = date
        self.stock_codes = stock_codes
        self.weights = weights
        if weights:
            if len(stock_codes) != len(weights):
                raise Exception('weights should have same length as stock_codes.')
        self.holding_period_in_month = holding_period_in_month
        self.omitted = list()
        self.label = label

    def __repr__(self):
        if self.label:
            return f"<{self.__class__.__name__} ({self.label})>"
        return f"<{self.__class__.__name__} ({self.id})>"



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader

from fontTools import ttLib, cffLib
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from fontTools.merge.base import add_method, mergeObjects
from fontTools.merge.cmap import computeMegaCmap
from fontTools.merge.util import *
import logging


log = logging.getLogger("fontTools.merge")


ttLib.getTableClass('maxp').mergeMap = {
	'*': max,
	'tableTag': equal,
	'tableVersion': equal,
	'numGlyphs': sum,
	'maxStorage': first,
	'maxFunctionDefs': first,
	'maxInstructionDefs': first,
	# TODO When we correctly merge hinting data, update these values:
	# maxFunctionDefs, maxInstructionDefs, maxSizeOfInstructions
}

headFlagsMergeBitMap = {
	'size': 16,
	'*': bitwise_or,
	1: bitwise_and, # Baseline at y = 0
	2: bitwise_and, # lsb at x = 0
	3: bitwise_and, # Force ppem to integer values. FIXME?
	5: bitwise_and, # Font is vertical
	6: lambda bit: 0, # Always set to zero
	11: bitwise_and, # Font data is 'lossless'
	13: bitwise_and, # Optimized for ClearType
	14: bitwise_and, # Last resort font. FIXME? equal or first may be better
	15: lambda bit: 0, # Always set to zero
}

ttLib.getTableClass('head').mergeMap = {
	'tableTag': equal,
	'tableVersion': max,
	'fontRevision': max,
	'checkSumAdjustment': lambda lst: 0, # We need *something* here
	'magicNumber': equal,
	'flags': mergeBits(headFlagsMergeBitMap),
	'unitsPerEm': equal,
	'created': current_time,
	'modified': current_time,
	'xMin': min,
	'yMin': min,
	'xMax': max,
	'yMax': max,
	'macStyle': first,
	'lowestRecPPEM': max,
	'fontDirectionHint': lambda lst: 2,
	'indexToLocFormat': first,
	'glyphDataFormat': equal,
}

ttLib.getTableClass('hhea').mergeMap = {
	'*': equal,
	'tableTag': equal,
	'tableVersion': max,
	'ascent': max,
	'descent': min,
	'lineGap': max,
	'advanceWidthMax': max,
	'minLeftSideBearing': min,
	'minRightSideBearing': min,
	'xMaxExtent': max,
	'caretSlopeRise': first,
	'caretSlopeRun': first,
	'caretOffset': first,
	'numberOfHMetrics': recalculate,
}

ttLib.getTableClass('vhea').mergeMap = {
	'*': equal,
	'tableTag': equal,
	'tableVersion': max,
	'ascent': max,
	'descent': min,
	'lineGap': max,
	'advanceHeightMax': max,
	'minTopSideBearing': min,
	'minBottomSideBearing': min,
	'yMaxExtent': max,
	'caretSlopeRise': first,
	'caretSlopeRun': first,
	'caretOffset': first,
	'numberOfVMetrics': recalculate,
}

os2FsTypeMergeBitMap = {
	'size': 16,
	'*': lambda bit: 0,
	1: bitwise_or, # no embedding permitted
	2: bitwise_and, # allow previewing and printing documents
	3: bitwise_and, # allow editing documents
	8: bitwise_or, # no subsetting permitted
	9: bitwise_or, # no embedding of outlines permitted
}

def mergeOs2FsType(lst):
	lst = list(lst)
	if all(item == 0 for item in lst):
		return 0

	# Compute least restrictive logic for each fsType value
	for i in range(len(lst)):
		# unset bit 1 (no embedding permitted) if either bit 2 or 3 is set
		if lst[i] & 0x000C:
			lst[i] &= ~0x0002
		# set bit 2 (allow previewing) if bit 3 is set (allow editing)
		elif lst[i] & 0x0008:
			lst[i] |= 0x0004
		# set bits 2 and 3 if everything is allowed
		elif lst[i] == 0:
			lst[i] = 0x000C

	fsType = mergeBits(os2FsTypeMergeBitMap)(lst)
	# unset bits 2 and 3 if bit 1 is set (some font is "no embedding")
	if fsType & 0x0002:
		fsType &= ~0x000C
	return fsType


ttLib.getTableClass('OS/2').mergeMap = {
	'*': first,
	'tableTag': equal,
	'version': max,
	'xAvgCharWidth': first, # Will be recalculated at the end on the merged font
	'fsType': mergeOs2FsType, # Will be overwritten
	'panose': first, # FIXME: should really be the first Latin font
	'ulUnicodeRange1': bitwise_or,
	'ulUnicodeRange2': bitwise_or,
	'ulUnicodeRange3': bitwise_or,
	'ulUnicodeRange4': bitwise_or,
	'fsFirstCharIndex': min,
	'fsLastCharIndex': max,
	'sTypoAscender': max,
	'sTypoDescender': min,
	'sTypoLineGap': max,
	'usWinAscent': max,
	'usWinDescent': max,
	# Version 1
	'ulCodePageRange1': onlyExisting(bitwise_or),
	'ulCodePageRange2': onlyExisting(bitwise_or),
	# Version 2, 3, 4
	'sxHeight': onlyExisting(max),
	'sCapHeight': onlyExisting(max),
	'usDefaultChar': onlyExisting(first),
	'usBreakChar': onlyExisting(first),
	'usMaxContext': onlyExisting(max),
	# version 5
	'usLowerOpticalPointSize': onlyExisting(min),
	'usUpperOpticalPointSize': onlyExisting(max),
}

@add_method(ttLib.getTableClass('OS/2'))
def merge(self, m, tables):
	DefaultTable.merge(self, m, tables)
	if self.version < 2:
		# bits 8 and 9 are reserved and should be set to zero
		self.fsType &= ~0x0300
	if self.version >= 3:
		# Only one of bits 1, 2, and 3 may be set. We already take
		# care of bit 1 implications in mergeOs2FsType. So unset
		# bit 2 if bit 3 is already set.
		if self.fsType & 0x0008:
			self.fsType &= ~0x0004
	return self

ttLib.getTableClass('post').mergeMap = {
	'*': first,
	'tableTag': equal,
	'formatType': max,
	'isFixedPitch': min,
	'minMemType42': max,
	'maxMemType42': lambda lst: 0,
	'minMemType1': max,
	'maxMemType1': lambda lst: 0,
	'mapping': onlyExisting(sumDicts),
	'extraNames': lambda lst: [],
}

ttLib.getTableClass('vmtx').mergeMap = ttLib.getTableClass('hmtx').mergeMap = {
	'tableTag': equal,
	'metrics': sumDicts,
}

ttLib.getTableClass('name').mergeMap = {
	'tableTag': equal,
	'names': first, # FIXME? Does mixing name records make sense?
}

ttLib.getTableClass('loca').mergeMap = {
	'*': recalculate,
	'tableTag': equal,
}

ttLib.getTableClass('glyf').mergeMap = {
	'tableTag': equal,
	'glyphs': sumDicts,
	'glyphOrder': sumLists,
}

@add_method(ttLib.getTableClass('glyf'))
def merge(self, m, tables):
	for i,table in enumerate(tables):
		for g in table.glyphs.values():
			if i:
				# Drop hints for all but first font, since
				# we don't map functions / CVT values.
				g.removeHinting()
			# Expand composite glyphs to load their
			# composite glyph names.
			if g.isComposite():
				g.expand(table)
	return DefaultTable.merge(self, m, tables)

ttLib.getTableClass('prep').mergeMap = lambda self, lst: first(lst)
ttLib.getTableClass('fpgm').mergeMap = lambda self, lst: first(lst)
ttLib.getTableClass('cvt ').mergeMap = lambda self, lst: first(lst)
ttLib.getTableClass('gasp').mergeMap = lambda self, lst: first(lst) # FIXME? Appears irreconcilable

@add_method(ttLib.getTableClass('CFF '))
def merge(self, m, tables):

	if any(hasattr(table, "FDSelect") for table in tables):
		raise NotImplementedError(
			"Merging CID-keyed CFF tables is not supported yet"
		)

	for table in tables:
		table.cff.desubroutinize()

	newcff = tables[0]
	newfont = newcff.cff[0]
	private = newfont.Private
	storedNamesStrings = []
	glyphOrderStrings = []
	glyphOrder = set(newfont.getGlyphOrder())

	for name in newfont.strings.strings:
		if name not in glyphOrder:
			storedNamesStrings.append(name)
		else:
			glyphOrderStrings.append(name)

	chrset = list(newfont.charset)
	newcs = newfont.CharStrings
	log.debug("FONT 0 CharStrings: %d.", len(newcs))

	for i, table in enumerate(tables[1:], start=1):
		font = table.cff[0]
		font.Private = private
		fontGlyphOrder = set(font.getGlyphOrder())
		for name in font.strings.strings:
			if name in fontGlyphOrder:
				glyphOrderStrings.append(name)
		cs = font.CharStrings
		gs = table.cff.GlobalSubrs
		log.debug("Font %d CharStrings: %d.", i, len(cs))
		chrset.extend(font.charset)
		if newcs.charStringsAreIndexed:
			for i, name in enumerate(cs.charStrings, start=len(newcs)):
				newcs.charStrings[name] = i
				newcs.charStringsIndex.items.append(None)
		for name in cs.charStrings:
			newcs[name] = cs[name]

	newfont.charset = chrset
	newfont.numGlyphs = len(chrset)
	newfont.strings.strings = glyphOrderStrings + storedNamesStrings

	return newcff

@add_method(ttLib.getTableClass('cmap'))
def merge(self, m, tables):

	# TODO Handle format=14.
	if not hasattr(m, 'cmap'):
		computeMegaCmap(m, tables)
	cmap = m.cmap

	cmapBmpOnly = {uni: gid for uni,gid in cmap.items() if uni <= 0xFFFF}
	self.tables = []
	module = ttLib.getTableModule('cmap')
	if len(cmapBmpOnly) != len(cmap):
		# format-12 required.
		cmapTable = module.cmap_classes[12](12)
		cmapTable.platformID = 3
		cmapTable.platEncID = 10
		cmapTable.language = 0
		cmapTable.cmap = cmap
		self.tables.append(cmapTable)
	# always create format-4
	cmapTable = module.cmap_classes[4](4)
	cmapTable.platformID = 3
	cmapTable.platEncID = 1
	cmapTable.language = 0
	cmapTable.cmap = cmapBmpOnly
	# ordered by platform then encoding
	self.tables.insert(0, cmapTable)
	self.tableVersion = 0
	self.numSubTables = len(self.tables)
	return self



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader

from fontTools.misc.timeTools import timestampNow
from fontTools.ttLib.tables.DefaultTable import DefaultTable
from functools import reduce
import operator
import logging


log = logging.getLogger("fontTools.merge")


# General utility functions for merging values from different fonts

def equal(lst):
	lst = list(lst)
	t = iter(lst)
	first = next(t)
	assert all(item == first for item in t), "Expected all items to be equal: %s" % lst
	return first

def first(lst):
	return next(iter(lst))

def recalculate(lst):
	return NotImplemented

def current_time(lst):
	return timestampNow()

def bitwise_and(lst):
	return reduce(operator.and_, lst)

def bitwise_or(lst):
	return reduce(operator.or_, lst)

def avg_int(lst):
	lst = list(lst)
	return sum(lst) // len(lst)

def onlyExisting(func):
	"""Returns a filter func that when called with a list,
	only calls func on the non-NotImplemented items of the list,
	and only so if there's at least one item remaining.
	Otherwise returns NotImplemented."""

	def wrapper(lst):
		items = [item for item in lst if item is not NotImplemented]
		return func(items) if items else NotImplemented

	return wrapper

def sumLists(lst):
	l = []
	for item in lst:
		l.extend(item)
	return l

def sumDicts(lst):
	d = {}
	for item in lst:
		d.update(item)
	return d

def mergeBits(bitmap):

	def wrapper(lst):
		lst = list(lst)
		returnValue = 0
		for bitNumber in range(bitmap['size']):
			try:
				mergeLogic = bitmap[bitNumber]
			except KeyError:
				try:
					mergeLogic = bitmap['*']
				except KeyError:
					raise Exception("Don't know how to merge bit %s" % bitNumber)
			shiftedBit = 1 << bitNumber
			mergedValue = mergeLogic(bool(item & shiftedBit) for item in lst)
			returnValue |= mergedValue << bitNumber
		return returnValue

	return wrapper


class AttendanceRecordingIdentityDict(object):
	"""A dictionary-like object that records indices of items actually accessed
	from a list."""

	def __init__(self, lst):
		self.l = lst
		self.d = {id(v):i for i,v in enumerate(lst)}
		self.s = set()

	def __getitem__(self, v):
		self.s.add(self.d[id(v)])
		return v

class GregariousIdentityDict(object):
	"""A dictionary-like object that welcomes guests without reservations and
	adds them to the end of the guest list."""

	def __init__(self, lst):
		self.l = lst
		self.s = set(id(v) for v in lst)

	def __getitem__(self, v):
		if id(v) not in self.s:
			self.s.add(id(v))
			self.l.append(v)
		return v

class NonhashableDict(object):
	"""A dictionary-like object mapping objects to values."""

	def __init__(self, keys, values=None):
		if values is None:
			self.d = {id(v):i for i,v in enumerate(keys)}
		else:
			self.d = {id(k):v for k,v in zip(keys, values)}

	def __getitem__(self, k):
		return self.d[id(k)]

	def __setitem__(self, k, v):
		self.d[id(k)] = v

	def __delitem__(self, k):
		del self.d[id(k)]



from .clients import *
from .tools.screener import Screener
from .tools.backtester import Backtester

def client(service_name: str):
    if service_name == 'variable':
        return VariableClient()
    elif service_name == 'stock_price':
        return StockPriceClient()

def tool(tool_name: str):
    if tool_name == 'screener':
        return Screener()
    elif tool_name == 'backtester':
        return Backtester()



from io import StringIO, BytesIO

import requests
import pandas as pd
import zipfile

class VariableClient:
    # opendata-finance-kr open api 서비스의 base url
    BASE_URL = 'http://apis.opendata-finance-kr.com'

    def __init__(self):
        self.variable_list = self.list_variables()

    def list_variables(self, to_dataframe=False):
        if to_dataframe:  # 데이터프레임으로 변환
            records = list()
            for var in self.variable_list:
                address = var.get('address')
                model_name, id = var.get('address').values()
                r = {
                    **var,
                    'model_name': address.get('model_name'),
                    'id': address.get('id')
                }
                address = r.pop('address')
                records.append(r)
            return pd.DataFrame.from_records(records)
        url = f"{self.BASE_URL}/variable/list"  # 요청 url 구성
        return self.request_data(url)           # 요청

    def request_data(self, url):
        res = requests.get(url)       # 요청 및 응답
        d = res.json()                # dictionary로 변환
        status = d.get('status')  # 응답 상태 확인
        if status['code'] == 204:
            return None
        return d.get('data')

    def get_cross_section(self, variable_name, yyyymm, to_dataframe=False):
        var = self.find_variable_by_name(variable_name)   # 지표 선택
        var_url = self.get_variable_request_url(var)     # 지표 요청 url 구성
        url = f"{var_url}/cross-section/{yyyymm}"         # 횡단면 데이터 요청 url 구성
        data = self.request_data(url)                     # 요청
        if not to_dataframe:
            return data
        return pd.DataFrame.from_records(data.get('records')) # 데이터프레임으로 변환

    def get_time_series(self, variable_name, stock_code, to_dataframe=False):
        var = self.find_variable_by_name(variable_name)  # 지표 선택
        var_url = self.get_variable_request_url(var)     # 지표 요청 url 구성
        url = f"{var_url}/time-series/{stock_code}"       # 시계열 데이터 요청 url 구성
        data = self.request_data(url)                     # 요청
        if not to_dataframe:
            return data
        return pd.DataFrame.from_records(data.get('records')) # 데이터프레임으로 변환

    def download_variable_panel(self, variable_name):
        var = self.find_variable_by_name(variable_name)  # 지표 선택
        requests.get(var['download_url'])                 # 다운로드
        return None


    def find_variable_by_name(self, name):
        filtered = list(filter(lambda var: var['name'] == name, self.variable_list))
        if len(filtered) == 0:
            return Exception(f"no matched variable named with {name}.")
        return filtered[0]

    def get_variable_request_url(self, variable):
        address = variable.get('address')
        return f"{self.BASE_URL}/variable/{address['model_name']}/{address['id']}"


class StockPriceClient:
    BASE_URL = 'http://apis.opendata-finance-kr.com'

    def __init__(self):
        self.data_list = self.list_data()

    def list_data(self, to_dataframe=False):
        if to_dataframe:
            return pd.DataFrame.from_records(self.data_list)
        url = f"{self.BASE_URL}/stock-price/list"
        return self.request_data(url)

    def request_data(self, url):
        res = requests.get(url)
        d = res.json()
        return d.get('data')

    def get_dataframe(self, date, column_name='en'):
        download_url = self.get_download_url(date)
        res = requests.get(download_url)
        zf = zipfile.ZipFile(BytesIO(res.content))
        fnm = zf.namelist()[0]
        with zf.open(fnm) as f:
            encoded = f.read()
            decoded = encoded.decode('utf-8')
        df = pd.read_csv(StringIO(decoded))
        df['basDt'] = df.basDt.astype(str)
        df = df[self.COLUMN_NAME_MAP.keys()] # ordering columns
        if column_name == 'raw':
            return df
        rnm_map = {
            k: v.get(column_name)
            for k, v in self.COLUMN_NAME_MAP.items()
        }
        return df.rename(columns=rnm_map)

    def get_download_url(self, date):
        ls = list(filter(
            lambda data: data['date'] == date,
            self.data_list
        ))
        if len(ls) == 0:
            raise Exception(f"No matched data with date {date}.")
        return ls[0].get('download_url')


    @property
    def COLUMN_NAME_MAP(self):
        return {
            'basDt': {'en': 'date', 'kr': '날짜'},
            'isinCd': {'en': 'isin_code', 'kr': 'ISIN코드'},
            'srtnCd': {'en': 'stock_code', 'kr': '종목코드'},
            'itmsNm': {'en': 'name', 'kr': '종목명'},
            'mrktCtg': {'en': 'market', 'kr': '시장구분'},
            'mkp': {'en': 'open', 'kr': '시가'},
            'hipr': {'en': 'high', 'kr': '고가'},
            'lopr': {'en': 'low', 'kr': '저가'},
            'clpr': {'en': 'close', 'kr': '종가'},
            'vs': {'en': 'delta_close', 'kr': '전일대비'},
            'fltRt': {'en': 'ri', 'kr': '수익률'},
            'lstgStCnt': {'en': 'n_listed', 'kr': '상장주식수'},
            'trqu': {'en': 'vol_n', 'kr': '거래량(주)'},
            'trPrc': {'en': 'vol_m', 'kr': '거래량(원)'},
            'mrktTotAmt': {'en': 'mktcap', 'kr': '시가총액'}
        }



"""GlyphSets returned by a TTFont."""

from abc import ABC, abstractmethod
from collections.abc import Mapping
from copy import copy
from fontTools.misc.fixedTools import otRound
from fontTools.misc.loggingTools import deprecateFunction


class _TTGlyphSet(Mapping):

    """Generic dict-like GlyphSet class that pulls metrics from hmtx and
    glyph shape from TrueType or CFF.
    """

    def __init__(self, font, location, glyphsMapping):
        self.font = font
        self.location = location
        self.glyphsMapping = glyphsMapping
        self.hMetrics = font["hmtx"].metrics
        self.vMetrics = getattr(font.get("vmtx"), "metrics", None)
        if location:
            from fontTools.varLib.varStore import VarStoreInstancer

            self.hvarTable = getattr(font.get("HVAR"), "table", None)
            if self.hvarTable is not None:
                self.hvarInstancer = VarStoreInstancer(
                    self.hvarTable.VarStore, font["fvar"].axes, location
                )
            # TODO VVAR, VORG

    def __contains__(self, glyphName):
        return glyphName in self.glyphsMapping

    def __iter__(self):
        return iter(self.glyphsMapping.keys())

    def __len__(self):
        return len(self.glyphsMapping)

    @deprecateFunction(
        "use 'glyphName in glyphSet' instead", category=DeprecationWarning
    )
    def has_key(self, glyphName):
        return glyphName in self.glyphsMapping


class _TTGlyphSetGlyf(_TTGlyphSet):
    def __init__(self, font, location):
        self.glyfTable = font["glyf"]
        super().__init__(font, location, self.glyfTable)
        if location:
            self.gvarTable = font.get("gvar")

    def __getitem__(self, glyphName):
        return _TTGlyphGlyf(self, glyphName)


class _TTGlyphSetCFF(_TTGlyphSet):
    def __init__(self, font, location):
        tableTag = "CFF2" if "CFF2" in font else "CFF "
        self.charStrings = list(font[tableTag].cff.values())[0].CharStrings
        super().__init__(font, location, self.charStrings)
        self.blender = None
        if location:
            from fontTools.varLib.varStore import VarStoreInstancer

            varStore = getattr(self.charStrings, "varStore", None)
            if varStore is not None:
                instancer = VarStoreInstancer(
                    varStore.otVarStore, font["fvar"].axes, location
                )
                self.blender = instancer.interpolateFromDeltas

    def __getitem__(self, glyphName):
        return _TTGlyphCFF(self, glyphName)


class _TTGlyph(ABC):

    """Glyph object that supports the Pen protocol, meaning that it has
    .draw() and .drawPoints() methods that take a pen object as their only
    argument. Additionally there are 'width' and 'lsb' attributes, read from
    the 'hmtx' table.

    If the font contains a 'vmtx' table, there will also be 'height' and 'tsb'
    attributes.
    """

    def __init__(self, glyphSet, glyphName):
        self.glyphSet = glyphSet
        self.name = glyphName
        self.width, self.lsb = glyphSet.hMetrics[glyphName]
        if glyphSet.vMetrics is not None:
            self.height, self.tsb = glyphSet.vMetrics[glyphName]
        else:
            self.height, self.tsb = None, None
        if glyphSet.location and glyphSet.hvarTable is not None:
            varidx = (
                glyphSet.font.getGlyphID(glyphName)
                if glyphSet.hvarTable.AdvWidthMap is None
                else glyphSet.hvarTable.AdvWidthMap.mapping[glyphName]
            )
            self.width += glyphSet.hvarInstancer[varidx]
        # TODO: VVAR/VORG

    @abstractmethod
    def draw(self, pen):
        """Draw the glyph onto ``pen``. See fontTools.pens.basePen for details
        how that works.
        """
        raise NotImplementedError

    def drawPoints(self, pen):
        """Draw the glyph onto ``pen``. See fontTools.pens.pointPen for details
        how that works.
        """
        from fontTools.pens.pointPen import SegmentToPointPen

        self.draw(SegmentToPointPen(pen))


class _TTGlyphGlyf(_TTGlyph):
    def draw(self, pen):
        """Draw the glyph onto ``pen``. See fontTools.pens.basePen for details
        how that works.
        """
        glyph, offset = self._getGlyphAndOffset()
        glyph.draw(pen, self.glyphSet.glyfTable, offset)

    def drawPoints(self, pen):
        """Draw the glyph onto ``pen``. See fontTools.pens.pointPen for details
        how that works.
        """
        glyph, offset = self._getGlyphAndOffset()
        glyph.drawPoints(pen, self.glyphSet.glyfTable, offset)

    def _getGlyphAndOffset(self):
        if self.glyphSet.location and self.glyphSet.gvarTable is not None:
            glyph = self._getGlyphInstance()
        else:
            glyph = self.glyphSet.glyfTable[self.name]

        offset = self.lsb - glyph.xMin if hasattr(glyph, "xMin") else 0
        return glyph, offset

    def _getGlyphInstance(self):
        from fontTools.varLib.iup import iup_delta
        from fontTools.ttLib.tables._g_l_y_f import GlyphCoordinates
        from fontTools.varLib.models import supportScalar

        glyphSet = self.glyphSet
        glyfTable = glyphSet.glyfTable
        variations = glyphSet.gvarTable.variations[self.name]
        hMetrics = glyphSet.hMetrics
        vMetrics = glyphSet.vMetrics
        coordinates, _ = glyfTable._getCoordinatesAndControls(
            self.name, hMetrics, vMetrics
        )
        origCoords, endPts = None, None
        for var in variations:
            scalar = supportScalar(glyphSet.location, var.axes)
            if not scalar:
                continue
            delta = var.coordinates
            if None in delta:
                if origCoords is None:
                    origCoords, control = glyfTable._getCoordinatesAndControls(
                        self.name, hMetrics, vMetrics
                    )
                    endPts = (
                        control[1] if control[0] >= 1 else list(range(len(control[1])))
                    )
                delta = iup_delta(delta, origCoords, endPts)
            coordinates += GlyphCoordinates(delta) * scalar

        glyph = copy(glyfTable[self.name])  # Shallow copy
        width, lsb, height, tsb = _setCoordinates(glyph, coordinates, glyfTable)
        self.lsb = lsb
        self.tsb = tsb
        if glyphSet.hvarTable is None:
            # no HVAR: let's set metrics from the phantom points
            self.width = width
            self.height = height
        return glyph


class _TTGlyphCFF(_TTGlyph):
    def draw(self, pen):
        """Draw the glyph onto ``pen``. See fontTools.pens.basePen for details
        how that works.
        """
        self.glyphSet.charStrings[self.name].draw(pen, self.glyphSet.blender)


def _setCoordinates(glyph, coord, glyfTable):
    # Handle phantom points for (left, right, top, bottom) positions.
    assert len(coord) >= 4
    leftSideX = coord[-4][0]
    rightSideX = coord[-3][0]
    topSideY = coord[-2][1]
    bottomSideY = coord[-1][1]

    for _ in range(4):
        del coord[-1]

    if glyph.isComposite():
        assert len(coord) == len(glyph.components)
        glyph.components = [copy(comp) for comp in glyph.components]  # Shallow copy
        for p, comp in zip(coord, glyph.components):
            if hasattr(comp, "x"):
                comp.x, comp.y = p
    elif glyph.numberOfContours == 0:
        assert len(coord) == 0
    else:
        assert len(coord) == len(glyph.coordinates)
        glyph.coordinates = coord

    glyph.recalcBounds(glyfTable)

    horizontalAdvanceWidth = otRound(rightSideX - leftSideX)
    verticalAdvanceWidth = otRound(topSideY - bottomSideY)
    leftSideBearing = otRound(glyph.xMin - leftSideX)
    topSideBearing = otRound(topSideY - glyph.yMax)
    return (
        horizontalAdvanceWidth,
        leftSideBearing,
        verticalAdvanceWidth,
        topSideBearing,
    )



from .types import ScreeningLogic
from functools import reduce
from typing import List
import pandas as pd
from opendata_finance_kr.clients import VariableClient

class Screener:
    variable_client = VariableClient()

    def __init__(self):
        self.logic_list = list()
        self.next_id = 1
        self.dataframe = pd.DataFrame()
        self.result = pd.DataFrame()

    def reset(self):
        return self.__init__()

    def create_logic(self, variable_name, yyyymm, quantile, keep):
        id = self.next_id
        self.next_id += 1
        logic = ScreeningLogic(
            id = id,
            variable_name = variable_name,
            yyyymm = yyyymm,
            quantile = quantile,
            keep = keep
        )
        self.logic_list.append(logic)
        return logic

    def get_logic(self, id):
        filtered = list(filter(lambda logic: logic.id == id, self.logic_list))
        if len(filtered) == 0:
            raise Exception(f'No logic with id {id}.')
        return filtered[0]

    def delete_logic(self, id):
        self.logic_list = list(filter(lambda logic: logic.id != id, self.logic_list))
        return None

    def order_logics(self, ordered_id: List[int]):
        self.logic_list = [self.get_logic(id) for id in ordered_id]
        return None

    def create_dataframe(self):
        if len(self.logic_list) == 0:
            return pd.DataFrame()
        ls_cs = list()
        for logic in self.logic_list:
            cs = self.variable_client.get_cross_section(
                variable_name = logic.variable_name,
                yyyymm = logic.yyyymm,
                to_dataframe = True
            )
            del cs['date']
            cs = cs.rename(columns={'value': logic.variable_name})
            cs = cs.set_index(['stock_code', 'market'])
            ls_cs.append(cs)
        df = pd.concat(ls_cs, axis=1)
        df = df.dropna().copy()
        self.dataframe = df
        return self.dataframe

    def execute(self, ordered):
        self.create_dataframe()
        if ordered:
            return self.execute_ordered()
        else:
            return self.execute_unordered()

    @property
    def METHOD_SET(self):
        return {'ordered', 'unordered'}

    def execute_ordered(self):
        screened = self.dataframe.copy()
        for logic in self.logic_list:
            s = screened[logic.variable_name]
            breakpoint = s.quantile(logic.quantile)
            if logic.keep == 'bottom':
                survivors = s <= breakpoint
            elif logic.keep == 'top':
                suvivors = s > breakpoint
            screened = screened[suvivors].copy()
        self.result = screened
        return self.result

    def execute_unordered(self):
        ls_survivors = list()
        for logic in self.logic_list:
            s = self.dataframe[logic.variable_name]
            breakpoint = s.quantile(logic.quantile)
            if logic.keep == 'bottom':
                ls_survivors.append(s <= breakpoint)
            elif logic.keep == 'top':
                ls_survivors.append(s > breakpoint)
            survivors = reduce(lambda x, y: x & y, ls_survivors)
        self.result = self.dataframe[survivors].copy()
        return self.result



from fontTools.pens.basePen import BasePen, OpenContourError
try:
	import cython
except ImportError:
	# if cython not installed, use mock module with no-op decorators and types
	from fontTools.misc import cython

if cython.compiled:
	# Yep, I'm compiled.
	COMPILED = True
else:
	# Just a lowly interpreted script.
	COMPILED = False


__all__ = ["MomentsPen"]

class MomentsPen(BasePen):

	def __init__(self, glyphset=None):
		BasePen.__init__(self, glyphset)

		self.area = 0
		self.momentX = 0
		self.momentY = 0
		self.momentXX = 0
		self.momentXY = 0
		self.momentYY = 0

	def _moveTo(self, p0):
		self.__startPoint = p0

	def _closePath(self):
		p0 = self._getCurrentPoint()
		if p0 != self.__startPoint:
			self._lineTo(self.__startPoint)

	def _endPath(self):
		p0 = self._getCurrentPoint()
		if p0 != self.__startPoint:
			# Green theorem is not defined on open contours.
			raise OpenContourError(
							"Green theorem is not defined on open contours."
			)

	@cython.locals(r0=cython.double)
	@cython.locals(r1=cython.double)
	@cython.locals(r2=cython.double)
	@cython.locals(r3=cython.double)
	@cython.locals(r4=cython.double)
	@cython.locals(r5=cython.double)
	@cython.locals(r6=cython.double)
	@cython.locals(r7=cython.double)
	@cython.locals(r8=cython.double)
	@cython.locals(r9=cython.double)
	@cython.locals(r10=cython.double)
	@cython.locals(r11=cython.double)
	@cython.locals(r12=cython.double)
	@cython.locals(x0=cython.double, y0=cython.double)
	@cython.locals(x1=cython.double, y1=cython.double)
	def _lineTo(self, p1):
		x0,y0 = self._getCurrentPoint()
		x1,y1 = p1

		r0 = x1*y0
		r1 = x1*y1
		r2 = x1**2
		r3 = r2*y1
		r4 = y0 - y1
		r5 = r4*x0
		r6 = x0**2
		r7 = 2*y0
		r8 = y0**2
		r9 = y1**2
		r10 = x1**3
		r11 = y0**3
		r12 = y1**3

		self.area += -r0/2 - r1/2 + x0*(y0 + y1)/2
		self.momentX += -r2*y0/6 - r3/3 - r5*x1/6 + r6*(r7 + y1)/6
		self.momentY += -r0*y1/6 - r8*x1/6 - r9*x1/6 + x0*(r8 + r9 + y0*y1)/6
		self.momentXX += -r10*y0/12 - r10*y1/4 - r2*r5/12 - r4*r6*x1/12 + x0**3*(3*y0 + y1)/12
		self.momentXY += -r2*r8/24 - r2*r9/8 - r3*r7/24 + r6*(r7*y1 + 3*r8 + r9)/24 - x0*x1*(r8 - r9)/12
		self.momentYY += -r0*r9/12 - r1*r8/12 - r11*x1/12 - r12*x1/12 + x0*(r11 + r12 + r8*y1 + r9*y0)/12

	@cython.locals(r0=cython.double)
	@cython.locals(r1=cython.double)
	@cython.locals(r2=cython.double)
	@cython.locals(r3=cython.double)
	@cython.locals(r4=cython.double)
	@cython.locals(r5=cython.double)
	@cython.locals(r6=cython.double)
	@cython.locals(r7=cython.double)
	@cython.locals(r8=cython.double)
	@cython.locals(r9=cython.double)
	@cython.locals(r10=cython.double)
	@cython.locals(r11=cython.double)
	@cython.locals(r12=cython.double)
	@cython.locals(r13=cython.double)
	@cython.locals(r14=cython.double)
	@cython.locals(r15=cython.double)
	@cython.locals(r16=cython.double)
	@cython.locals(r17=cython.double)
	@cython.locals(r18=cython.double)
	@cython.locals(r19=cython.double)
	@cython.locals(r20=cython.double)
	@cython.locals(r21=cython.double)
	@cython.locals(r22=cython.double)
	@cython.locals(r23=cython.double)
	@cython.locals(r24=cython.double)
	@cython.locals(r25=cython.double)
	@cython.locals(r26=cython.double)
	@cython.locals(r27=cython.double)
	@cython.locals(r28=cython.double)
	@cython.locals(r29=cython.double)
	@cython.locals(r30=cython.double)
	@cython.locals(r31=cython.double)
	@cython.locals(r32=cython.double)
	@cython.locals(r33=cython.double)
	@cython.locals(r34=cython.double)
	@cython.locals(r35=cython.double)
	@cython.locals(r36=cython.double)
	@cython.locals(r37=cython.double)
	@cython.locals(r38=cython.double)
	@cython.locals(r39=cython.double)
	@cython.locals(r40=cython.double)
	@cython.locals(r41=cython.double)
	@cython.locals(r42=cython.double)
	@cython.locals(r43=cython.double)
	@cython.locals(r44=cython.double)
	@cython.locals(r45=cython.double)
	@cython.locals(r46=cython.double)
	@cython.locals(r47=cython.double)
	@cython.locals(r48=cython.double)
	@cython.locals(r49=cython.double)
	@cython.locals(r50=cython.double)
	@cython.locals(r51=cython.double)
	@cython.locals(r52=cython.double)
	@cython.locals(r53=cython.double)
	@cython.locals(x0=cython.double, y0=cython.double)
	@cython.locals(x1=cython.double, y1=cython.double)
	@cython.locals(x2=cython.double, y2=cython.double)
	def _qCurveToOne(self, p1, p2):
		x0,y0 = self._getCurrentPoint()
		x1,y1 = p1
		x2,y2 = p2

		r0 = 2*y1
		r1 = r0*x2
		r2 = x2*y2
		r3 = 3*r2
		r4 = 2*x1
		r5 = 3*y0
		r6 = x1**2
		r7 = x2**2
		r8 = 4*y1
		r9 = 10*y2
		r10 = 2*y2
		r11 = r4*x2
		r12 = x0**2
		r13 = 10*y0
		r14 = r4*y2
		r15 = x2*y0
		r16 = 4*x1
		r17 = r0*x1 + r2
		r18 = r2*r8
		r19 = y1**2
		r20 = 2*r19
		r21 = y2**2
		r22 = r21*x2
		r23 = 5*r22
		r24 = y0**2
		r25 = y0*y2
		r26 = 5*r24
		r27 = x1**3
		r28 = x2**3
		r29 = 30*y1
		r30 = 6*y1
		r31 = 10*r7*x1
		r32 = 5*y2
		r33 = 12*r6
		r34 = 30*x1
		r35 = x1*y1
		r36 = r3 + 20*r35
		r37 = 12*x1
		r38 = 20*r6
		r39 = 8*r6*y1
		r40 = r32*r7
		r41 = 60*y1
		r42 = 20*r19
		r43 = 4*r19
		r44 = 15*r21
		r45 = 12*x2
		r46 = 12*y2
		r47 = 6*x1
		r48 = 8*r19*x1 + r23
		r49 = 8*y1**3
		r50 = y2**3
		r51 = y0**3
		r52 = 10*y1
		r53 = 12*y1

		self.area += -r1/6 - r3/6 + x0*(r0 + r5 + y2)/6 + x1*y2/3 - y0*(r4 + x2)/6
		self.momentX += -r11*(-r10 + y1)/30 + r12*(r13 + r8 + y2)/30 + r6*y2/15 - r7*r8/30 - r7*r9/30 + x0*(r14 - r15 - r16*y0 + r17)/30 - y0*(r11 + 2*r6 + r7)/30
		self.momentY += -r18/30 - r20*x2/30 - r23/30 - r24*(r16 + x2)/30 + x0*(r0*y2 + r20 + r21 + r25 + r26 + r8*y0)/30 + x1*y2*(r10 + y1)/15 - y0*(r1 + r17)/30
		self.momentXX += r12*(r1 - 5*r15 - r34*y0 + r36 + r9*x1)/420 + 2*r27*y2/105 - r28*r29/420 - r28*y2/4 - r31*(r0 - 3*y2)/420 - r6*x2*(r0 - r32)/105 + x0**3*(r30 + 21*y0 + y2)/84 - x0*(r0*r7 + r15*r37 - r2*r37 - r33*y2 + r38*y0 - r39 - r40 + r5*r7)/420 - y0*(8*r27 + 5*r28 + r31 + r33*x2)/420
		self.momentXY += r12*(r13*y2 + 3*r21 + 105*r24 + r41*y0 + r42 + r46*y1)/840 - r16*x2*(r43 - r44)/840 - r21*r7/8 - r24*(r38 + r45*x1 + 3*r7)/840 - r41*r7*y2/840 - r42*r7/840 + r6*y2*(r32 + r8)/210 + x0*(-r15*r8 + r16*r25 + r18 + r21*r47 - r24*r34 - r26*x2 + r35*r46 + r48)/420 - y0*(r16*r2 + r30*r7 + r35*r45 + r39 + r40)/420
		self.momentYY += -r2*r42/420 - r22*r29/420 - r24*(r14 + r36 + r52*x2)/420 - r49*x2/420 - r50*x2/12 - r51*(r47 + x2)/84 + x0*(r19*r46 + r21*r5 + r21*r52 + r24*r29 + r25*r53 + r26*y2 + r42*y0 + r49 + 5*r50 + 35*r51)/420 + x1*y2*(r43 + r44 + r9*y1)/210 - y0*(r19*r45 + r2*r53 - r21*r4 + r48)/420

	@cython.locals(r0=cython.double)
	@cython.locals(r1=cython.double)
	@cython.locals(r2=cython.double)
	@cython.locals(r3=cython.double)
	@cython.locals(r4=cython.double)
	@cython.locals(r5=cython.double)
	@cython.locals(r6=cython.double)
	@cython.locals(r7=cython.double)
	@cython.locals(r8=cython.double)
	@cython.locals(r9=cython.double)
	@cython.locals(r10=cython.double)
	@cython.locals(r11=cython.double)
	@cython.locals(r12=cython.double)
	@cython.locals(r13=cython.double)
	@cython.locals(r14=cython.double)
	@cython.locals(r15=cython.double)
	@cython.locals(r16=cython.double)
	@cython.locals(r17=cython.double)
	@cython.locals(r18=cython.double)
	@cython.locals(r19=cython.double)
	@cython.locals(r20=cython.double)
	@cython.locals(r21=cython.double)
	@cython.locals(r22=cython.double)
	@cython.locals(r23=cython.double)
	@cython.locals(r24=cython.double)
	@cython.locals(r25=cython.double)
	@cython.locals(r26=cython.double)
	@cython.locals(r27=cython.double)
	@cython.locals(r28=cython.double)
	@cython.locals(r29=cython.double)
	@cython.locals(r30=cython.double)
	@cython.locals(r31=cython.double)
	@cython.locals(r32=cython.double)
	@cython.locals(r33=cython.double)
	@cython.locals(r34=cython.double)
	@cython.locals(r35=cython.double)
	@cython.locals(r36=cython.double)
	@cython.locals(r37=cython.double)
	@cython.locals(r38=cython.double)
	@cython.locals(r39=cython.double)
	@cython.locals(r40=cython.double)
	@cython.locals(r41=cython.double)
	@cython.locals(r42=cython.double)
	@cython.locals(r43=cython.double)
	@cython.locals(r44=cython.double)
	@cython.locals(r45=cython.double)
	@cython.locals(r46=cython.double)
	@cython.locals(r47=cython.double)
	@cython.locals(r48=cython.double)
	@cython.locals(r49=cython.double)
	@cython.locals(r50=cython.double)
	@cython.locals(r51=cython.double)
	@cython.locals(r52=cython.double)
	@cython.locals(r53=cython.double)
	@cython.locals(r54=cython.double)
	@cython.locals(r55=cython.double)
	@cython.locals(r56=cython.double)
	@cython.locals(r57=cython.double)
	@cython.locals(r58=cython.double)
	@cython.locals(r59=cython.double)
	@cython.locals(r60=cython.double)
	@cython.locals(r61=cython.double)
	@cython.locals(r62=cython.double)
	@cython.locals(r63=cython.double)
	@cython.locals(r64=cython.double)
	@cython.locals(r65=cython.double)
	@cython.locals(r66=cython.double)
	@cython.locals(r67=cython.double)
	@cython.locals(r68=cython.double)
	@cython.locals(r69=cython.double)
	@cython.locals(r70=cython.double)
	@cython.locals(r71=cython.double)
	@cython.locals(r72=cython.double)
	@cython.locals(r73=cython.double)
	@cython.locals(r74=cython.double)
	@cython.locals(r75=cython.double)
	@cython.locals(r76=cython.double)
	@cython.locals(r77=cython.double)
	@cython.locals(r78=cython.double)
	@cython.locals(r79=cython.double)
	@cython.locals(r80=cython.double)
	@cython.locals(r81=cython.double)
	@cython.locals(r82=cython.double)
	@cython.locals(r83=cython.double)
	@cython.locals(r84=cython.double)
	@cython.locals(r85=cython.double)
	@cython.locals(r86=cython.double)
	@cython.locals(r87=cython.double)
	@cython.locals(r88=cython.double)
	@cython.locals(r89=cython.double)
	@cython.locals(r90=cython.double)
	@cython.locals(r91=cython.double)
	@cython.locals(r92=cython.double)
	@cython.locals(r93=cython.double)
	@cython.locals(r94=cython.double)
	@cython.locals(r95=cython.double)
	@cython.locals(r96=cython.double)
	@cython.locals(r97=cython.double)
	@cython.locals(r98=cython.double)
	@cython.locals(r99=cython.double)
	@cython.locals(r100=cython.double)
	@cython.locals(r101=cython.double)
	@cython.locals(r102=cython.double)
	@cython.locals(r103=cython.double)
	@cython.locals(r104=cython.double)
	@cython.locals(r105=cython.double)
	@cython.locals(r106=cython.double)
	@cython.locals(r107=cython.double)
	@cython.locals(r108=cython.double)
	@cython.locals(r109=cython.double)
	@cython.locals(r110=cython.double)
	@cython.locals(r111=cython.double)
	@cython.locals(r112=cython.double)
	@cython.locals(r113=cython.double)
	@cython.locals(r114=cython.double)
	@cython.locals(r115=cython.double)
	@cython.locals(r116=cython.double)
	@cython.locals(r117=cython.double)
	@cython.locals(r118=cython.double)
	@cython.locals(r119=cython.double)
	@cython.locals(r120=cython.double)
	@cython.locals(r121=cython.double)
	@cython.locals(r122=cython.double)
	@cython.locals(r123=cython.double)
	@cython.locals(r124=cython.double)
	@cython.locals(r125=cython.double)
	@cython.locals(r126=cython.double)
	@cython.locals(r127=cython.double)
	@cython.locals(r128=cython.double)
	@cython.locals(r129=cython.double)
	@cython.locals(r130=cython.double)
	@cython.locals(r131=cython.double)
	@cython.locals(r132=cython.double)
	@cython.locals(x0=cython.double, y0=cython.double)
	@cython.locals(x1=cython.double, y1=cython.double)
	@cython.locals(x2=cython.double, y2=cython.double)
	@cython.locals(x3=cython.double, y3=cython.double)
	def _curveToOne(self, p1, p2, p3):
		x0,y0 = self._getCurrentPoint()
		x1,y1 = p1
		x2,y2 = p2
		x3,y3 = p3

		r0 = 6*y2
		r1 = r0*x3
		r2 = 10*y3
		r3 = r2*x3
		r4 = 3*y1
		r5 = 6*x1
		r6 = 3*x2
		r7 = 6*y1
		r8 = 3*y2
		r9 = x2**2
		r10 = 45*r9
		r11 = r10*y3
		r12 = x3**2
		r13 = r12*y2
		r14 = r12*y3
		r15 = 7*y3
		r16 = 15*x3
		r17 = r16*x2
		r18 = x1**2
		r19 = 9*r18
		r20 = x0**2
		r21 = 21*y1
		r22 = 9*r9
		r23 = r7*x3
		r24 = 9*y2
		r25 = r24*x2 + r3
		r26 = 9*x2
		r27 = x2*y3
		r28 = -r26*y1 + 15*r27
		r29 = 3*x1
		r30 = 45*x1
		r31 = 12*x3
		r32 = 45*r18
		r33 = 5*r12
		r34 = r8*x3
		r35 = 105*y0
		r36 = 30*y0
		r37 = r36*x2
		r38 = 5*x3
		r39 = 15*y3
		r40 = 5*y3
		r41 = r40*x3
		r42 = x2*y2
		r43 = 18*r42
		r44 = 45*y1
		r45 = r41 + r43 + r44*x1
		r46 = y2*y3
		r47 = r46*x3
		r48 = y2**2
		r49 = 45*r48
		r50 = r49*x3
		r51 = y3**2
		r52 = r51*x3
		r53 = y1**2
		r54 = 9*r53
		r55 = y0**2
		r56 = 21*x1
		r57 = 6*x2
		r58 = r16*y2
		r59 = r39*y2
		r60 = 9*r48
		r61 = r6*y3
		r62 = 3*y3
		r63 = r36*y2
		r64 = y1*y3
		r65 = 45*r53
		r66 = 5*r51
		r67 = x2**3
		r68 = x3**3
		r69 = 630*y2
		r70 = 126*x3
		r71 = x1**3
		r72 = 126*x2
		r73 = 63*r9
		r74 = r73*x3
		r75 = r15*x3 + 15*r42
		r76 = 630*x1
		r77 = 14*x3
		r78 = 21*r27
		r79 = 42*x1
		r80 = 42*x2
		r81 = x1*y2
		r82 = 63*r42
		r83 = x1*y1
		r84 = r41 + r82 + 378*r83
		r85 = x2*x3
		r86 = r85*y1
		r87 = r27*x3
		r88 = 27*r9
		r89 = r88*y2
		r90 = 42*r14
		r91 = 90*x1
		r92 = 189*r18
		r93 = 378*r18
		r94 = r12*y1
		r95 = 252*x1*x2
		r96 = r79*x3
		r97 = 30*r85
		r98 = r83*x3
		r99 = 30*x3
		r100 = 42*x3
		r101 = r42*x1
		r102 = r10*y2 + 14*r14 + 126*r18*y1 + r81*r99
		r103 = 378*r48
		r104 = 18*y1
		r105 = r104*y2
		r106 = y0*y1
		r107 = 252*y2
		r108 = r107*y0
		r109 = y0*y3
		r110 = 42*r64
		r111 = 378*r53
		r112 = 63*r48
		r113 = 27*x2
		r114 = r27*y2
		r115 = r113*r48 + 42*r52
		r116 = x3*y3
		r117 = 54*r42
		r118 = r51*x1
		r119 = r51*x2
		r120 = r48*x1
		r121 = 21*x3
		r122 = r64*x1
		r123 = r81*y3
		r124 = 30*r27*y1 + r49*x2 + 14*r52 + 126*r53*x1
		r125 = y2**3
		r126 = y3**3
		r127 = y1**3
		r128 = y0**3
		r129 = r51*y2
		r130 = r112*y3 + r21*r51
		r131 = 189*r53
		r132 = 90*y2

		self.area += -r1/20 - r3/20 - r4*(x2 + x3)/20 + x0*(r7 + r8 + 10*y0 + y3)/20 + 3*x1*(y2 + y3)/20 + 3*x2*y3/10 - y0*(r5 + r6 + x3)/20
		self.momentX += r11/840 - r13/8 - r14/3 - r17*(-r15 + r8)/840 + r19*(r8 + 2*y3)/840 + r20*(r0 + r21 + 56*y0 + y3)/168 + r29*(-r23 + r25 + r28)/840 - r4*(10*r12 + r17 + r22)/840 + x0*(12*r27 + r30*y2 + r34 - r35*x1 - r37 - r38*y0 + r39*x1 - r4*x3 + r45)/840 - y0*(r17 + r30*x2 + r31*x1 + r32 + r33 + 18*r9)/840
		self.momentY += -r4*(r25 + r58)/840 - r47/8 - r50/840 - r52/6 - r54*(r6 + 2*x3)/840 - r55*(r56 + r57 + x3)/168 + x0*(r35*y1 + r40*y0 + r44*y2 + 18*r48 + 140*r55 + r59 + r63 + 12*r64 + r65 + r66)/840 + x1*(r24*y1 + 10*r51 + r59 + r60 + r7*y3)/280 + x2*y3*(r15 + r8)/56 - y0*(r16*y1 + r31*y2 + r44*x2 + r45 + r61 - r62*x1)/840
		self.momentXX += -r12*r72*(-r40 + r8)/9240 + 3*r18*(r28 + r34 - r38*y1 + r75)/3080 + r20*(r24*x3 - r72*y0 - r76*y0 - r77*y0 + r78 + r79*y3 + r80*y1 + 210*r81 + r84)/9240 - r29*(r12*r21 + 14*r13 + r44*r9 - r73*y3 + 54*r86 - 84*r87 - r89 - r90)/9240 - r4*(70*r12*x2 + 27*r67 + 42*r68 + r74)/9240 + 3*r67*y3/220 - r68*r69/9240 - r68*y3/4 - r70*r9*(-r62 + y2)/9240 + 3*r71*(r24 + r40)/3080 + x0**3*(r24 + r44 + 165*y0 + y3)/660 + x0*(r100*r27 + 162*r101 + r102 + r11 + 63*r18*y3 + r27*r91 - r33*y0 - r37*x3 + r43*x3 - r73*y0 - r88*y1 + r92*y2 - r93*y0 - 9*r94 - r95*y0 - r96*y0 - r97*y1 - 18*r98 + r99*x1*y3)/9240 - y0*(r12*r56 + r12*r80 + r32*x3 + 45*r67 + 14*r68 + 126*r71 + r74 + r85*r91 + 135*r9*x1 + r92*x2)/9240
		self.momentXY += -r103*r12/18480 - r12*r51/8 - 3*r14*y2/44 + 3*r18*(r105 + r2*y1 + 18*r46 + 15*r48 + 7*r51)/6160 + r20*(1260*r106 + r107*y1 + r108 + 28*r109 + r110 + r111 + r112 + 30*r46 + 2310*r55 + r66)/18480 - r54*(7*r12 + 18*r85 + 15*r9)/18480 - r55*(r33 + r73 + r93 + r95 + r96 + r97)/18480 - r7*(42*r13 + r82*x3 + 28*r87 + r89 + r90)/18480 - 3*r85*(r48 - r66)/220 + 3*r9*y3*(r62 + 2*y2)/440 + x0*(-r1*y0 - 84*r106*x2 + r109*r56 + 54*r114 + r117*y1 + 15*r118 + 21*r119 + 81*r120 + r121*r46 + 54*r122 + 60*r123 + r124 - r21*x3*y0 + r23*y3 - r54*x3 - r55*r72 - r55*r76 - r55*r77 + r57*y0*y3 + r60*x3 + 84*r81*y0 + 189*r81*y1)/9240 + x1*(r104*r27 - r105*x3 - r113*r53 + 63*r114 + r115 - r16*r53 + 28*r47 + r51*r80)/3080 - y0*(54*r101 + r102 + r116*r5 + r117*x3 + 21*r13 - r19*y3 + r22*y3 + r78*x3 + 189*r83*x2 + 60*r86 + 81*r9*y1 + 15*r94 + 54*r98)/9240
		self.momentYY += -r103*r116/9240 - r125*r70/9240 - r126*x3/12 - 3*r127*(r26 + r38)/3080 - r128*(r26 + r30 + x3)/660 - r4*(r112*x3 + r115 - 14*r119 + 84*r47)/9240 - r52*r69/9240 - r54*(r58 + r61 + r75)/9240 - r55*(r100*y1 + r121*y2 + r26*y3 + r79*y2 + r84 + 210*x2*y1)/9240 + x0*(r108*y1 + r110*y0 + r111*y0 + r112*y0 + 45*r125 + 14*r126 + 126*r127 + 770*r128 + 42*r129 + r130 + r131*y2 + r132*r64 + 135*r48*y1 + 630*r55*y1 + 126*r55*y2 + 14*r55*y3 + r63*y3 + r65*y3 + r66*y0)/9240 + x1*(27*r125 + 42*r126 + 70*r129 + r130 + r39*r53 + r44*r48 + 27*r53*y2 + 54*r64*y2)/3080 + 3*x2*y3*(r48 + r66 + r8*y3)/220 - y0*(r100*r46 + 18*r114 - 9*r118 - 27*r120 - 18*r122 - 30*r123 + r124 + r131*x2 + r132*x3*y1 + 162*r42*y1 + r50 + 63*r53*x3 + r64*r99)/9240

if __name__ == '__main__':
	from fontTools.misc.symfont import x, y, printGreenPen
	printGreenPen('MomentsPen', [
		      ('area', 1),
		      ('momentX', x),
		      ('momentY', y),
		      ('momentXX', x**2),
		      ('momentXY', x*y),
		      ('momentYY', y**2),
		     ])



from .types import Portfolio
from dateutil.relativedelta import relativedelta
from opendata_finance_kr.clients import StockPriceClient
from warnings import warn
import datetime
import pandas as pd


class Backtester:
    stock_price_client = StockPriceClient()
    format_date = lambda dt: datetime.datetime.strptime(dt, '%Y%m%d').date()

    def __init__(self):
        self.date_range = self.range_date()
        self.portfolio_list: List[Portfolio] = list()
        self.dataframe = pd.DataFrame()
        self.next_id = 1
        self.portfolio_list_changed = False
        self.query = dict()
        self.query_changed = False
        self.result = pd.DataFrame()

    def reset(self):
        return self.__init__()

    def range_date(self):
        return [
            self.stock_price_client.data_list[0]['date'],
            self.stock_price_client.data_list[-1]['date']
        ]

    def create_portfolio(self, date, stock_codes, weights=None, holding_period_in_month=None, label=None):
        pf = Portfolio(
            id = self.next_id,
            date = date,
            stock_codes = stock_codes,
            weights = weights,
            holding_period_in_month = holding_period_in_month,
            label = label
        )
        validated = self.validate_portfolio(pf)
        self.portfolio_list.append(validated)
        self.next_id += 1
        self.portfolio_list_changed = True
        return validated

    def validate_portfolio(self, portfolio: Portfolio):
        labels = [pf.label for pf in self.portfolio_list]
        if portfolio.label in labels:
            raise Exception(f"Portfolio labelled by '{portfolio.label}' has already exists.")
        if portfolio.date < self.date_range[0]:
            portfolio.date = self.date_range[0]
            warn(f"Input date is replaced to {self.date_range[0]} because it is lesser than the min date of service.")
        df = self.stock_price_client.get_dataframe(portfolio.date)
        omitted = [
            sc for sc in portfolio.stock_codes
            if df.stock_code.isin([sc]).empty
        ]
        if len(omitted) > 0:
            portfolio.omitted += omitted
            warn(f"{len(omitted)} number of stock codes were omitted because price data missing.")

        # some validation process for weights and holding_period_in_month

        # append date range to arrange query
        if not portfolio.holding_period_in_month:
            to_date = self.date_range[-1]
        else:
            to_date = format_date(portfolio.date) + relativedelta(months=portfolio.holding_period_in_month)
            to_date = to_date.strftime('%Y%m%d')
        portfolio.date_range = [portfolio.date, to_date]
        return portfolio

    def get_portfolio(self, id_or_label):
        k = self.identify_id_or_label(id_or_label)
        filtered = list(filter(
            lambda pf: getattr(pf, k) == id_or_label,
            self.portfolio_list
        ))
        if len(filtered) == 0:
            raise Exception(f"No matched portfolio with {k} {id_or_label}.")
        return filtered[0]

    def delete_portfolio(self, id_or_label):
        k = self.identify_id_or_label(id_or_label)
        self.portfolio_list = list(filter(
            lambda pf: getattr(pf, k) != id_or_label,
            self.portfolio_list
        ))
        self.portfolio_list_changed = True
        return None

    def identify_id_or_label(self, id_or_label):
        if type(id_or_label) == int:
            return 'id'
        elif type(id_or_label) == str:
            return 'label'
        else:
            raise Exception('Invalid id or label.')

    def create_dataframe(self, return_dataframe=False):
        if len(self.portfolio_list) == 0:
            raise Exception('No portfolio to create dataframe. Please create portfolio.')
        if not self.query_changed:
            return self.dataframe
        for date, stock_code_set in self.query.items():
            df = self.stock_price_client.get_dataframe(date)
            self.dataframe = pd.concat(
                [self.dataframe, df.loc[df.stock_code.isin(stock_code_set)]],
                axis = 0,
                ignore_index = True
            )
        self.portfolio_list_changed = False
        if return_dataframe:
            return self.dataframe
        return None

    def create_query(self, return_query=False):
        if not self.portfolio_list_changed:
            return self.query

        # set date range to query
        min_date = min([pf.date_range[0] for pf in self.portfolio_list])
        max_date = max([pf.date_range[1] for pf in self.portfolio_list])
        query = {
            data['date']: list()
            for data in self.stock_price_client.data_list
            if (data['date'] >= min_date) and (data['date'] <= max_date)
        }

        # set stock codes to query
        for date in query.keys():
            for pf in self.portfolio_list:
                from_date, to_date = pf.date_range
                matched = (date >= from_date) and (date <= to_date)
                if not matched:
                    continue
                query[date] += pf.stock_codes

        # drop duplicates
        for date in query.keys():
            stock_code_set = sorted(list(set(query[date])))
            query[date] = stock_code_set

        # update query
        self.query_changed = self.query != query
        if self.query_changed:
            self.query = query

        if return_query:
            return self.query
        return None

    def execute(self):
        result = pd.DataFrame()
        if self.portfolio_list_changed:
            self.create_query()
            if self.query_changed:
                # return self.dataframe
                print('Creating dataframe for backtest... It takes a couple of minutes.')
                self.create_dataframe()
        for pf in self.portfolio_list:
            result_pf = self.execute_by_portfolio(pf)
            result = pd.concat([result, result_pf], axis=0, ignore_index=True)
        self.result = result.set_index(['date', 'portfolio']).unstack('portfolio')
        return self.result

    def execute_by_portfolio(self, portfolio):
        # subset
        df = self.dataframe[
            (self.dataframe.stock_code.isin(portfolio.stock_codes))
            & (self.dataframe.date >= portfolio.date_range[0])
            & (self.dataframe.date <= portfolio.date_range[1])
        ]
        if portfolio.weights:
            d = dict(zip(portfolio.stock_codes, portfolio.weights))
            w_records = [{
                'stock_code': stock_code,
                'weight': weight
            } for stock_code, weight in d.items()]
            dfw = pd.DataFrame.from_records(w_records)
        else:
            dfw = df[df.date == portfolio.date]
            dfw = dfw[['stock_code', 'date', 'mktcap']]
            sum_mktcap = dfw.mktcap.sum()
            dfw['weight'] = dfw.mktcap / sum_mktcap
            dfw = dfw[['stock_code', 'weight']]
        df = df.merge(dfw, on=['stock_code'])
        df = df.sort_values(['stock_code', 'date'])
        Ri = df.ri / 100 + 1
        is_first = df.stock_code != df.stock_code.shift(1)
        Ri.loc[is_first] = 1
        cum_Ri = Ri.groupby(df.stock_code).cumprod()
        w_cum_Ri = df.weight * cum_Ri

        result_pf = w_cum_Ri.groupby(df.date).sum().rename('cumrp').reset_index()
        result_pf['rp'] = result_pf['cumrp'] / result_pf['cumrp'].shift(1)

        R2pct = lambda s: ((s - 1)*100).round(2)
        result_pf.cumrp = R2pct(result_pf.cumrp)
        result_pf.rp = R2pct(result_pf.rp)
        result_pf['portfolio'] = portfolio.id if not portfolio.label else portfolio.label
        return result_pf[['date', 'portfolio', 'rp', 'cumrp']]

    def create_report(self):
        return



"""Compiles/decompiles SVG table.

https://docs.microsoft.com/en-us/typography/opentype/spec/svg

The XML format is:

.. code-block:: xml

	<SVG>
		<svgDoc endGlyphID="1" startGlyphID="1">
			<![CDATA[ <complete SVG doc> ]]
		</svgDoc>
	...
		<svgDoc endGlyphID="n" startGlyphID="m">
			<![CDATA[ <complete SVG doc> ]]
		</svgDoc>
	</SVG>
"""

from fontTools.misc.textTools import bytesjoin, safeEval, strjoin, tobytes, tostr
from fontTools.misc import sstruct
from . import DefaultTable
from collections.abc import Sequence
from dataclasses import dataclass, astuple
from io import BytesIO
import struct
import logging


log = logging.getLogger(__name__)


SVG_format_0 = """
	>   # big endian
	version:                  H
	offsetToSVGDocIndex:      L
	reserved:                 L
"""

SVG_format_0Size = sstruct.calcsize(SVG_format_0)

doc_index_entry_format_0 = """
	>   # big endian
	startGlyphID:             H
	endGlyphID:               H
	svgDocOffset:             L
	svgDocLength:             L
"""

doc_index_entry_format_0Size = sstruct.calcsize(doc_index_entry_format_0)



class table_S_V_G_(DefaultTable.DefaultTable):

	def decompile(self, data, ttFont):
		self.docList = []
		# Version 0 is the standardized version of the table; and current.
		# https://www.microsoft.com/typography/otspec/svg.htm
		sstruct.unpack(SVG_format_0, data[:SVG_format_0Size], self)
		if self.version != 0:
			log.warning(
				"Unknown SVG table version '%s'. Decompiling as version 0.", self.version)
		# read in SVG Documents Index
		# data starts with the first entry of the entry list.
		pos = subTableStart = self.offsetToSVGDocIndex
		self.numEntries = struct.unpack(">H", data[pos:pos+2])[0]
		pos += 2
		if self.numEntries > 0:
			data2 = data[pos:]
			entries = []
			for i in range(self.numEntries):
				docIndexEntry, data2 = sstruct.unpack2(doc_index_entry_format_0, data2, DocumentIndexEntry())
				entries.append(docIndexEntry)

			for entry in entries:
				start = entry.svgDocOffset + subTableStart
				end = start + entry.svgDocLength
				doc = data[start:end]
				compressed = False
				if doc.startswith(b"\x1f\x8b"):
					import gzip
					bytesIO = BytesIO(doc)
					with gzip.GzipFile(None, "r", fileobj=bytesIO) as gunzipper:
						doc = gunzipper.read()
					del bytesIO
					compressed = True
				doc = tostr(doc, "utf_8")
				self.docList.append(
					SVGDocument(doc, entry.startGlyphID, entry.endGlyphID, compressed)
				)

	def compile(self, ttFont):
		version = 0
		offsetToSVGDocIndex = SVG_format_0Size # I start the SVGDocIndex right after the header.
		# get SGVDoc info.
		docList = []
		entryList = []
		numEntries = len(self.docList)
		datum = struct.pack(">H",numEntries)
		entryList.append(datum)
		curOffset = len(datum) + doc_index_entry_format_0Size*numEntries
		seenDocs = {}
		allCompressed = getattr(self, "compressed", False)
		for i, doc in enumerate(self.docList):
			if isinstance(doc, (list, tuple)):
				doc = SVGDocument(*doc)
				self.docList[i] = doc
			docBytes = tobytes(doc.data, encoding="utf_8")
			if (allCompressed or doc.compressed) and not docBytes.startswith(b"\x1f\x8b"):
				import gzip
				bytesIO = BytesIO()
				# mtime=0 strips the useless timestamp and makes gzip output reproducible;
				# equivalent to `gzip -n`
				with gzip.GzipFile(None, "w", fileobj=bytesIO, mtime=0) as gzipper:
					gzipper.write(docBytes)
				gzipped = bytesIO.getvalue()
				if len(gzipped) < len(docBytes):
					docBytes = gzipped
				del gzipped, bytesIO
			docLength = len(docBytes)
			if docBytes in seenDocs:
				docOffset = seenDocs[docBytes]
			else:
				docOffset = curOffset
				curOffset += docLength
				seenDocs[docBytes] = docOffset
				docList.append(docBytes)
			entry = struct.pack(">HHLL", doc.startGlyphID, doc.endGlyphID, docOffset, docLength)
			entryList.append(entry)
		entryList.extend(docList)
		svgDocData = bytesjoin(entryList)

		reserved = 0
		header = struct.pack(">HLL", version, offsetToSVGDocIndex, reserved)
		data = [header, svgDocData]
		data = bytesjoin(data)
		return data

	def toXML(self, writer, ttFont):
		for i, doc in enumerate(self.docList):
			if isinstance(doc, (list, tuple)):
				doc = SVGDocument(*doc)
				self.docList[i] = doc
			attrs = {"startGlyphID": doc.startGlyphID, "endGlyphID": doc.endGlyphID}
			if doc.compressed:
				attrs["compressed"] = 1
			writer.begintag("svgDoc", **attrs)
			writer.newline()
			writer.writecdata(doc.data)
			writer.newline()
			writer.endtag("svgDoc")
			writer.newline()

	def fromXML(self, name, attrs, content, ttFont):
		if name == "svgDoc":
			if not hasattr(self, "docList"):
				self.docList = []
			doc = strjoin(content)
			doc = doc.strip()
			startGID = int(attrs["startGlyphID"])
			endGID = int(attrs["endGlyphID"])
			compressed = bool(safeEval(attrs.get("compressed", "0")))
			self.docList.append(SVGDocument(doc, startGID, endGID, compressed))
		else:
			log.warning("Unknown %s %s", name, content)


class DocumentIndexEntry(object):
	def __init__(self):
		self.startGlyphID = None # USHORT
		self.endGlyphID = None # USHORT
		self.svgDocOffset = None # ULONG
		self.svgDocLength = None # ULONG

	def __repr__(self):
		return "startGlyphID: %s, endGlyphID: %s, svgDocOffset: %s, svgDocLength: %s" % (self.startGlyphID, self.endGlyphID, self.svgDocOffset, self.svgDocLength)


@dataclass
class SVGDocument(Sequence):
	data: str
	startGlyphID: int
	endGlyphID: int
	compressed: bool = False

	# Previously, the SVG table's docList attribute contained a lists of 3 items:
	# [doc, startGlyphID, endGlyphID]; later, we added a `compressed` attribute.
	# For backward compatibility with code that depends of them being sequences of
	# fixed length=3, we subclass the Sequence abstract base class and pretend only
	# the first three items are present. 'compressed' is only accessible via named
	# attribute lookup like regular dataclasses: i.e. `doc.compressed`, not `doc[3]`
	def __getitem__(self, index):
		return astuple(self)[:3][index]

	def __len__(self):
		return 3



"""Benchmark the cu2qu algorithm performance."""

from .cu2qu import *
import random
import timeit

MAX_ERR = 5

def generate_curve():
    return [
        tuple(float(random.randint(0, 2048)) for coord in range(2))
        for point in range(4)]

def setup_curve_to_quadratic():
    return generate_curve(), MAX_ERR

def setup_curves_to_quadratic():
    num_curves = 3
    return (
        [generate_curve() for curve in range(num_curves)],
        [MAX_ERR] * num_curves)

def run_benchmark(
        benchmark_module, module, function, setup_suffix='', repeat=5, number=1000):
    setup_func = 'setup_' + function
    if setup_suffix:
        print('%s with %s:' % (function, setup_suffix), end='')
        setup_func += '_' + setup_suffix
    else:
        print('%s:' % function, end='')

    def wrapper(function, setup_func):
        function = globals()[function]
        setup_func = globals()[setup_func]
        def wrapped():
            return function(*setup_func())
        return wrapped
    results = timeit.repeat(wrapper(function, setup_func), repeat=repeat, number=number)
    print('\t%5.1fus' % (min(results) * 1000000. / number))

def main():
    """Benchmark the cu2qu algorithm performance."""
    run_benchmark('cu2qu.benchmark', 'cu2qu', 'curve_to_quadratic')
    run_benchmark('cu2qu.benchmark', 'cu2qu', 'curves_to_quadratic')


if __name__ == '__main__':
    random.seed(1)
    main()



from .otBase import BaseTTXConverter


class table__f_e_a_t(BaseTTXConverter):
	"""The feature name table is an AAT (Apple Advanced Typography) table for
	storing font features, settings, and their human-readable names. It should
	not be confused with the ``Feat`` table or the OpenType Layout ``GSUB``/``GPOS``
	tables. See `Feature Name Table <https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6feat.html>`_
	in the TrueType Reference Manual for more information on the structure and
	purpose of this table."""
	pass



"""fontTools.ttLib -- a package for dealing with TrueType fonts."""

from fontTools.misc.loggingTools import deprecateFunction
import logging


log = logging.getLogger(__name__)

class TTLibError(Exception): pass

@deprecateFunction("use logging instead", category=DeprecationWarning)
def debugmsg(msg):
	import time
	print(msg + time.strftime("  (%H:%M:%S)", time.localtime(time.time())))

from fontTools.ttLib.ttFont import *
from fontTools.ttLib.ttCollection import TTCollection



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader

from fontTools.merge.unicode import is_Default_Ignorable
from fontTools.pens.recordingPen import DecomposingRecordingPen
import logging


log = logging.getLogger("fontTools.merge")


def computeMegaGlyphOrder(merger, glyphOrders):
	"""Modifies passed-in glyphOrders to reflect new glyph names.
    Stores merger.glyphOrder."""
	megaOrder = {}
	for glyphOrder in glyphOrders:
		for i,glyphName in enumerate(glyphOrder):
			if glyphName in megaOrder:
				n = megaOrder[glyphName]
				while (glyphName + "." + repr(n)) in megaOrder:
					n += 1
				megaOrder[glyphName] = n
				glyphName += "." + repr(n)
				glyphOrder[i] = glyphName
			megaOrder[glyphName] = 1
	merger.glyphOrder = megaOrder = list(megaOrder.keys())


def _glyphsAreSame(glyphSet1, glyphSet2, glyph1, glyph2,
				   advanceTolerance=.05,
				   advanceToleranceEmpty=.20):
	pen1 = DecomposingRecordingPen(glyphSet1)
	pen2 = DecomposingRecordingPen(glyphSet2)
	g1 = glyphSet1[glyph1]
	g2 = glyphSet2[glyph2]
	g1.draw(pen1)
	g2.draw(pen2)
	if pen1.value != pen2.value:
		return False
	# Allow more width tolerance for glyphs with no ink
	tolerance = advanceTolerance if pen1.value else advanceToleranceEmpty
    # TODO Warn if advances not the same but within tolerance.
	if abs(g1.width - g2.width) > g1.width * tolerance:
		return False
	if hasattr(g1, 'height') and g1.height is not None:
		if abs(g1.height - g2.height) > g1.height * tolerance:
			return False
	return True

# Valid (format, platformID, platEncID) triplets for cmap subtables containing
# Unicode BMP-only and Unicode Full Repertoire semantics.
# Cf. OpenType spec for "Platform specific encodings":
# https://docs.microsoft.com/en-us/typography/opentype/spec/name
class _CmapUnicodePlatEncodings:
	BMP = {(4, 3, 1), (4, 0, 3), (4, 0, 4), (4, 0, 6)}
	FullRepertoire = {(12, 3, 10), (12, 0, 4), (12, 0, 6)}

def computeMegaCmap(merger, cmapTables):
	"""Sets merger.cmap and merger.glyphOrder."""

	# TODO Handle format=14.
	# Only merge format 4 and 12 Unicode subtables, ignores all other subtables
	# If there is a format 12 table for a font, ignore the format 4 table of it
	chosenCmapTables = []
	for fontIdx,table in enumerate(cmapTables):
		format4 = None
		format12 = None
		for subtable in table.tables:
			properties = (subtable.format, subtable.platformID, subtable.platEncID)
			if properties in _CmapUnicodePlatEncodings.BMP:
				format4 = subtable
			elif properties in _CmapUnicodePlatEncodings.FullRepertoire:
				format12 = subtable
			else:
				log.warning(
					"Dropped cmap subtable from font '%s':\t"
					"format %2s, platformID %2s, platEncID %2s",
					fontIdx, subtable.format, subtable.platformID, subtable.platEncID
				)
		if format12 is not None:
			chosenCmapTables.append((format12, fontIdx))
		elif format4 is not None:
			chosenCmapTables.append((format4, fontIdx))

	# Build the unicode mapping
	merger.cmap = cmap = {}
	fontIndexForGlyph = {}
	glyphSets = [None for f in merger.fonts] if hasattr(merger, 'fonts') else None

	for table,fontIdx in chosenCmapTables:
		# handle duplicates
		for uni,gid in table.cmap.items():
			oldgid = cmap.get(uni, None)
			if oldgid is None:
				cmap[uni] = gid
				fontIndexForGlyph[gid] = fontIdx
			elif is_Default_Ignorable(uni) or uni in (0x25CC,): # U+25CC DOTTED CIRCLE
				continue
			elif oldgid != gid:
				# Char previously mapped to oldgid, now to gid.
				# Record, to fix up in GSUB 'locl' later.
				if merger.duplicateGlyphsPerFont[fontIdx].get(oldgid) is None:
					if glyphSets is not None:
						oldFontIdx = fontIndexForGlyph[oldgid]
						for idx in (fontIdx, oldFontIdx):
							if glyphSets[idx] is None:
								glyphSets[idx] = merger.fonts[idx].getGlyphSet()
						#if _glyphsAreSame(glyphSets[oldFontIdx], glyphSets[fontIdx], oldgid, gid):
						#	continue
					merger.duplicateGlyphsPerFont[fontIdx][oldgid] = gid
				elif merger.duplicateGlyphsPerFont[fontIdx][oldgid] != gid:
					# Char previously mapped to oldgid but oldgid is already remapped to a different
					# gid, because of another Unicode character.
					# TODO: Try harder to do something about these.
					log.warning("Dropped mapping from codepoint %#06X to glyphId '%s'", uni, gid)


def renameCFFCharStrings(merger, glyphOrder, cffTable):
	"""Rename topDictIndex charStrings based on glyphOrder."""
	td = cffTable.cff.topDictIndex[0]

	charStrings = {}
	for i, v in enumerate(td.CharStrings.charStrings.values()):
		glyphName = glyphOrder[i]
		charStrings[glyphName] = v
	td.CharStrings.charStrings = charStrings

	td.charset = list(glyphOrder)



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Behdad Esfahbod, Roozbeh Pournader

from fontTools.ttLib.tables.DefaultTable import DefaultTable
import logging


log = logging.getLogger("fontTools.merge")


def add_method(*clazzes, **kwargs):
	"""Returns a decorator function that adds a new method to one or
	more classes."""
	allowDefault = kwargs.get('allowDefaultTable', False)
	def wrapper(method):
		done = []
		for clazz in clazzes:
			if clazz in done: continue # Support multiple names of a clazz
			done.append(clazz)
			assert allowDefault or clazz != DefaultTable, 'Oops, table class not found.'
			assert method.__name__ not in clazz.__dict__, \
				"Oops, class '%s' has method '%s'." % (clazz.__name__, method.__name__)
			setattr(clazz, method.__name__, method)
		return None
	return wrapper

def mergeObjects(lst):
	lst = [item for item in lst if item is not NotImplemented]
	if not lst:
		return NotImplemented
	lst = [item for item in lst if item is not None]
	if not lst:
		return None

	clazz = lst[0].__class__
	assert all(type(item) == clazz for item in lst), lst

	logic = clazz.mergeMap
	returnTable = clazz()
	returnDict = {}

	allKeys = set.union(set(), *(vars(table).keys() for table in lst))
	for key in allKeys:
		try:
			mergeLogic = logic[key]
		except KeyError:
			try:
				mergeLogic = logic['*']
			except KeyError:
				raise Exception("Don't know how to merge key %s of class %s" %
						(key, clazz.__name__))
		if mergeLogic is NotImplemented:
			continue
		value = mergeLogic(getattr(table, key, NotImplemented) for table in lst)
		if value is not NotImplemented:
			returnDict[key] = value

	returnTable.__dict__ = returnDict

	return returnTable

@add_method(DefaultTable, allowDefaultTable=True)
def merge(self, m, tables):
	if not hasattr(self, 'mergeMap'):
		log.info("Don't know how to merge '%s'.", self.tableTag)
		return NotImplemented

	logic = self.mergeMap

	if isinstance(logic, dict):
		return m.mergeObjects(self, self.mergeMap, tables)
	else:
		return logic(tables)


